/***************************************************************************
    file	         : rk_dcoplib.cpp
    copyright            : (C)	1999,2000,2001,2002,2003,2004,2005
				 by Mike Richardson
			   (C)	2001,2002,2003,2004,2005
				by John Dean
    license              : This file is released under the terms of
                           the GNU General Public License, version 2. The
                           copyright holders retain the right to release
                           this code under diffenent non-exclusive licences.
    email                : mike@quaking.demon.co.uk                                     
 ***************************************************************************/


#include	<qdatastream.h>
#include	<qregexp.h>
#include	<qcstring.h>
#include	<qvariant.h>

#if		defined(Q_OS_LINUX) || defined(Q_OS_MACX)
#include	<stdlib.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<string.h>
#include	<unistd.h>
#endif

#if		defined(Q_OS_WIN32)
#include	<winsock.h>
#endif

#include	<errno.h>

#include	"librkcop_exports.h"

typedef	struct sockaddr_in		SOCKIN	;
typedef	struct sockaddr			SOCKADR	;


#if		defined(Q_OS_LINUX)
#define		SOCKET			int
#define		SOCKETCLOSE(s)		close(s)
#define		SOCKETWRITE(s,b,l)	write(s,b,l)
#define		SOCKETREAD(s,b,l)	read(s,b,l)
#endif
#if		defined(Q_OS_MACX)
#define		SOCKET			int
#define		SOCKETCLOSE(s)		close(s)
#define		SOCKETWRITE(s,b,l)	write(s,b,l)
#define		SOCKETREAD(s,b,l)	read(s,b,l)
#endif
#if		defined(Q_OS_WIN32)
#define		SOCKETCLOSE(s)		closesocket(s)
#define		SOCKETWRITE(s,b,l)	send(s,b,l,0)
#define		SOCKETREAD(s,b,l)	recv(s,b,l,0)
#endif



/*  rkDCOPOpen	: Open RKDCOP connection				*/
/*  host	: const char *	: Host name (currently IP address)	*/
/*  sock	: SOCKET &	: Return socket on success		*/
/*  (returns)	: bool		: Success				*/

static	bool	rkDCOPOpen
	(	const char		*host,
		SOCKET			&sock
	)
{
#if	defined(Q_OS_WIN32)
	WSADATA wsa ;
	WSAStartup  (MAKEWORD(1,1), &wsa) ;
#endif
	SOCKIN	addr	;
	memset 	(&addr, 0, sizeof(addr))	;
	addr.sin_family      = AF_INET		;
	addr.sin_port	     = htons (6990)	;

	QRegExp	ae ("([0-9]+)\\.([0-9]+)\\.([0-9]+)\\.([0-9]+)") ;
	if (ae.search (host) >= 0)
	{
		long	ip	=
			((ae.cap(1).toInt() & 0xff) << 24) |
			((ae.cap(2).toInt() & 0xff) << 16) |
			((ae.cap(3).toInt() & 0xff) <<  8) |
			((ae.cap(4).toInt() & 0xff) <<  0) ;

		addr.sin_addr.s_addr = htonl (ip) ;
	}
	else	addr.sin_addr.s_addr = htonl ((127<<24)|1) ;

	sock = socket (PF_INET, SOCK_STREAM, 0) ;
	if (connect (sock, (SOCKADR *)&addr, sizeof(addr)) == 0)
		return	true	;

	SOCKETCLOSE(sock) ;
	return	false	  ;
}

/*  rkDCOPCall	: Make RKDCOP call					*/
/*  host	: const char *	: Host name (currently IP address)	*/
/*  target	: const char *	: Target object in application		*/
/*  func	: const char *	: Function to call in target object	*/
/*  args	: QValueList<QCString>					*/
/*				: Arguments to function			*/
/*  resv	: QVariant &	: Return value				*/
/*  (returns)	: QString	: Error message or null if OK		*/

QString	LIBRKCOP_API
	rkDCOPCall
	(	const char		*host,
		const char		*target,
		const char		*func,
		QValueList<QCString>	&args,
		QVariant		&resv
	)
{
	/* Extract the argument specifications from the function	*/
	/* specification, and then check that we have the right number	*/
	/* of arguments on the command line.				*/
	QRegExp	re ("\\((.*)\\)") ;
	if (re.search (func) < 0)
		return	"RKDCOP: Invalid function" ;

	QStringList types	= QStringList::split (',', re.cap(1)) ;
	if (args.count() != types.count())
		return	"RKDCOP: missing or extra arguments" ;


	/* Marshall the arguments, according to type, into a byte array	*/
	/* by way of a data stream. This ensures platform independance	*/
	/* and is in any case needed since that is what KDE's DCOP uses	*/
	/* Mark the end with an additional zero value.			*/
	QByteArray	data	;
	QDataStream	stream	(data, IO_WriteOnly) ;
	for (uint idx = 0 ; idx < types.count() ; idx += 1)
	{
		if (types[idx] == "QString")
		{	stream	<< QString(args[idx]) ;
			continue   ;
		}
		if (types[idx] == "int")
		{	stream	<< args[idx].toInt()  ;
			continue   ;
		}

		return	QString("RKDCOP: %1: unknown type").arg(types[idx]) ;
	}

	stream	<< 0 ;

	/* Open a socket to the application ...				*/
	SOCKET	sock	;
	if (!rkDCOPOpen (host, sock))
		return	"RKDCOP: cannot connect" ;

	/* The wire data starts with a null-terminated header which	*/
	/* contains the size of the following data, plus the target	*/
	/* object name and the function specification.			*/
	char	b[1024]	;
	sprintf	    (b, "%d|%s|%s", data.size(), target, func) ;
	if (SOCKETWRITE (sock, b, strlen(b) + 1) != (int)(strlen(b) + 1))
		return	"RKDCOP: write error" ;
	if (SOCKETWRITE (sock, data.data(), data.size()) != (int)data.size())
		return	"RKDCOP: write error" ;

	/* Now read back the response. We loop until we have both the	*/
	/* reply header (which contains the reply length and the reply	*/
	/* data type) and the reply data.				*/
	QByteArray reply ;
	char	   buff	 [1024] ;
	int	   got	 ;
	int	   zero	 ;

	while ((got = SOCKETREAD (sock, buff, sizeof(buff))) > 0)
	{
		int now = reply.size() ;
		reply.resize (now + got)   ;
		memcpy (reply.data() + now, buff, got) ;

		if ((zero = reply.find (0)) < 0)
			continue ;
		if ((int)reply.size() >= zero + 1 + atoi(reply.data()))
			break	 ;
	}

	SOCKETCLOSE (sock) ;

	if (reply.data() == 0)
		return	"RKDCOP: lost connection to rekall" ;


	/* Extract the reply type from the header, and get a data 	*/
	/* stream which contains the data *after* the header. We can	*/
	/* then extract the reply data according to the reply type.	*/
	const char *type  = strchr (reply.data(), '|') + 1 ;
	QByteArray  ba ;
	ba.duplicate (reply.data() + zero + 1, reply.size() - zero - 1) ;
	QDataStream result (ba, IO_ReadOnly) ;

	if (strcmp (type, "#Error") == 0)
	{
		return	"RKDCOP: processing error" ;
	}

	if (strcmp (type, "QStringList") == 0)
	{
		QStringList rv ;
		result   >> rv ;
		resv	 =  rv ;
		return	QString::null ;
	}

	if (strcmp (type, "QString"    ) == 0)
	{
		QString     rv ;
		result	 >> rv ;
		resv	 =  rv ;
		return	QString::null ;
	}

	if (strcmp (type, "int"        ) == 0)
	{
		int         rv ;
		result	 >> rv ;
		resv	 =  rv ;
		return	QString::null ;
	}

	if (strcmp (type, "bool"       ) == 0)
	{
		int         rv ;
		result	 >> rv ;
		resv	 =  QVariant::QVariant((bool)rv, 0) ;
		return	QString::null ;
	}

	return	QString("RKDCOP: %1: unknown result type").arg(type) ;
}

/*  rkDCOPAvailable							*/
/*		: See if RKDCOP channel is available			*/
/*  host	: const char *	: Host name (currently IP address)	*/
/*  (returns)	: bool		: True if available			*/

bool	LIBRKCOP_API
	rkDCOPAvailable
	(	const char		*host
	)
{

	SOCKET	sock	;
	if (!rkDCOPOpen (host, sock))
		return	false	;

	SOCKETCLOSE(sock) ;
	return	true;
}
